-- macros
	
	macroScript real_isoLevels
		category:"# Scripts"
		tooltip:"Isolation Levels - New Iso"
		buttontext:"New Iso"
	(
		on execute do with undo off if cIsoLevels != undefined then cIsoLevels.newIso()
	)
	
	macroScript real_isoLevels_exitIso
		category:"# Scripts"
		tooltip:"Isolation Levels - Exit Iso"
		buttontext:"Exit Iso"
	(
		on isEnabled do cIsoLevels != undefined and cIsoLevels.bIsoMode
		on execute do with undo off cIsoLevels.exitIso()
	)
	
	macroScript real_isoLevels_isoForward
		category:"# Scripts"
		tooltip:"Isolation Levels - Forward"
		buttontext:"Iso. Levels Forward"
		icon:#("IsoLevels2",1)
	(
		on isEnabled do cIsoLevels != undefined and cIsoLevels.lIsoVis[cIsoLevels.nIsoLevel+1] != undefined and cIsoLevels.bIsoMode
		on execute do with undo off cIsoLevels.isoForward()
	)
	
	macroScript real_isoLevels_isoBackward
		category:"# Scripts"
		tooltip:"Isolation Levels - Backward"
		buttontext:"Iso. Levels Backward"
		icon:#("IsoLevels1",1)
	(
		on isEnabled do cIsoLevels != undefined and cIsoLevels.nIsoLevel > 0
		on execute do with undo off cIsoLevels.isoBackward()
	)

-- script
(
	global cIsoLevels = rollout cIsoLevels ""
	(
		-- vars
			
			local lHiddenNoIsoObjs = #()
			local lVisNoIsoObjs = #()
			local lIsoObjs = #()
			
			local bIsoMode = false
			local lXrefVis = #()
			local lXrefChildsVis = #()
			local lIsoVis = #()
			
			local nIsoLevel = 0
			local tRandom = "isoLevels213120471205"
			local cmdPanelModeBackup = #create
		
		-- functions
			
			fn sceneNodeAddedCallback =
			(
				if cIsoLevels.bIsoMode then
				(
					local obj = callbacks.notificationParam() 
					
					if obj.isHidden
						then append lHiddenNoIsoObjs obj
						else append lIsoObjs obj
				)
			)
			
			fn sceneNodeDeletedCallback =
			(
				if cIsoLevels.bIsoMode then
				(
					local obj = callbacks.notificationParam()
					
					local nIndex
					
					if (nIndex = findItem lIsoObjs obj) != 0 then deleteItem lIsoObjs nIndex
						else 
					if (nIndex = findItem lVisNoIsoObjs obj) != 0 then deleteItem lVisNoIsoObjs nIndex
						else
					if (nIndex = findItem lHiddenNoIsoObjs obj) != 0 then deleteItem lHiddenNoIsoObjs nIndex
				)
			)
			
			fn getObj nHandle =
			(
				for obj in objects where obj.inode.handle == nHandle do return obj
			)
			
			fn exitIso =
			(
				try unhide lVisNoIsoObjs catch unhide (for obj in lVisNoIsoObjs where obj != undefined and not isDeleted obj collect obj)
				try unhide lIsoObjs catch unhide (for obj in lIsoObjs where obj != undefined and not isDeleted obj collect obj)
				cIsoLevels.setXrefsVis true
				
				nIsoLevel = 0
				bIsoMode = false
				cIsoLevels.updateUi()
			)
			
			fn updateUi =
			(
				local pMaxHwnd = windows.getMaxHWND()
				local tTitle = UIAccessor.GetWindowText pMaxHwnd
				local nAutodesk = findstring tTitle " - Autodesk"
				tTitle = substring tTitle 1 (nAutodesk-1)
				if matchPattern tTitle pattern:"* - ISO [*/*]" then
				(
					lTitle = filterstring tTitle " "
					tTitle = ""
					for n=1 to lTitle.count-3 do tTitle += lTitle[n] +" "
					tTitle = substring tTitle 1 (tTitle.count-1)
				)
				if bIsoMode then tTitle += " - ISO ["+ nIsoLevel as string +"/"+ lIsoVis.count as string +"]" 
				cui.setAppTitle tTitle
				
				redrawViews()
			)
			
			fn disableXrefRend parent =
			(
				local lRend = #()
				
				for nChild=1 to parent.children.count do
				(
					local child = parent.children[nChild]
					append lRend child.renderable
					append lRend child.ishidden
					child.renderable = false
					child.ishidden = true
					
					if child.children.count > 0 then join lRend (cIsoLevels.disableXrefRend child)
				)
				
				return lRend
			)
			
			fn restoreXrefRend parent n1 n2:1 =
			(
				for nChild=1 to parent.children.count do
				(
					local child = parent.children[nChild]
					child.renderable = lXrefChildsVis[n1][n2]
					n2 += 1
					child.ishidden = lXrefChildsVis[n1][n2]
					n2 += 1
					
					if child.children.count > 0 then n2 = cIsoLevels.restoreXrefRend child n1 n2:n2
				)
				
				return n2
			)
			
			fn setXrefsVis bRend =
			(
				if xrefs.getXRefFileCount() > 0 then
				(
					if not bRend then
					(
						lXrefVis = #()
						lXrefChildsVis = #()
						
						for nXref=1 to xrefs.getXRefFileCount() do
						(
							local xref = xrefs.getXRefFile nXref
							if doesFileExist xref.filename and xref.tree.children.count > 0 then
							(
								setuserprop xref.tree.children[1] tRandom nXref -- tag first child in xref
								lXrefChildsVis[nXref] = cIsoLevels.disableXrefRend xref.tree
								lXrefVis[nXref] = xref.hidden
								xref.hidden = true
							)
						)
					)
					else
					(
						for nXref=1 to xrefs.getXRefFileCount() do
						(
							local xref = xrefs.getXRefFile nXref
							if doesFileExist xref.filename and xref.tree.children.count > 0 then
							(
								local nIndex = getuserprop xref.tree.children[1] tRandom
								if nIndex != undefined then
								(
									cIsoLevels.restoreXrefRend xref.tree nIndex
									xref.hidden = lXrefVis[nIndex]
								)
							)
						)
					)
				)
			)
			
			fn optimizeIso b =
			(
				if b == on then
				(
					cmdPanelModeBackup = getCommandPanelTaskMode()
					setCommandPanelTaskMode mode:#modify
				)
				else
				(
					setCommandPanelTaskMode mode:cmdPanelModeBackup
				)
			)
			
			fn isoForward =
			(
				cIsoLevels.optimizeIso on
				
				nIsoLevel += 1
				
				local lHideObjs = #()
				for nObj=lIsoObjs.count to 1 by -1 do
				(
					local obj = lIsoObjs[nObj]
					if obj == undefined or isDeleted obj then deleteItem lIsoObjs nObj
					else if not obj.isHidden and lIsoVis[nIsoLevel][obj.inode.handle] then append lHideObjs obj
				)
				hide lHideObjs
				
				cIsoLevels.updateUi()
				
				if nIsoLevel == 1 then
				(
					try hide lVisNoIsoObjs catch
					(
						lVisNoIsoObjs = for obj in lVisNoIsoObjs where obj != undefined and not isDeleted obj collect obj
						hide lVisNoIsoObjs
					)
					cIsoLevels.setXrefsVis false
				)
				
				cIsoLevels.optimizeIso off
			)
			
			fn isoBackward =
			(
				cIsoLevels.optimizeIso on
				
				local lUnhideObjs = #()
				for nObj=lIsoObjs.count to 1 by -1 do
				(
					local obj = lIsoObjs[nObj]
					if obj == undefined or isDeleted obj then deleteItem lIsoObjs nObj
					else if obj.isHidden and lIsoVis[nIsoLevel][obj.inode.handle] then append lUnhideObjs obj
				)
				unhide lUnhideObjs
				
				nIsoLevel -= 1
				
				cIsoLevels.updateUi()
				
				if nIsoLevel == 0 then
				(
					unhide lVisNoIsoObjs
					cIsoLevels.setXrefsVis true
				)
				
				cIsoLevels.optimizeIso off
			)
			
			fn newIso =
			(
				if selection.count > 0 and selection.count != objects.count then
				(
					cIsoLevels.optimizeIso on
					
					bIsoMode = true
					nIsoLevel += 1
					lIsoVis[nIsoLevel] = #{}
					
					-- cleanup no longer needed isolevels
					for n=lIsoVis.count to nIsoLevel+1 by -1 do deleteItem lIsoVis n
					
					if nIsoLevel == 1 then
					(
						lHiddenNoIsoObjs = #()
						lVisNoIsoObjs = #()
						lIsoObjs = #()
						local lTargetObjs = #()
						
						for obj in objects where findItem lTargetObjs obj == 0 do
						(
							-- hidden objects
							if obj.isHidden
								then append lHiddenNoIsoObjs obj
								else 
									-- visible, selected objects
									if obj.isSelected
										then
										(
											append lIsoObjs obj
											
											-- include targets of selected objects
											if isproperty obj "target" and obj.target != undefined then
											(
												append lTargetObjs obj.target
												append lIsoObjs obj.target
											)
										)
										else
											-- visible, unselected objects
											append lVisNoIsoObjs obj
						)
						
						hide lVisNoIsoObjs
						cIsoLevels.setXrefsVis false
					)
					else
					(
						-- hide
							
							local lHideObjs = #()
							for obj in lIsoObjs where not obj.isSelected and not obj.isHidden do
							(
								lIsoVis[nIsoLevel][obj.inode.handle] = true
								append lHideObjs obj
							)
							
							if lHideObjs.count > 0 then
							(
								hide lHideObjs
								for n=lIsoVis.count to (nIsoLevel+1) by -1 do deleteItem lIsoVis n
								cIsoLevels.updateUi()
							)
							else
							(
								deleteItem lIsoVis nIsoLevel
								nIsoLevel -= 1
							)
					)
					
					cIsoLevels.updateUi()
					cIsoLevels.optimizeIso off
				)
			)
	)
)

-- callbacks

	callbacks.addScript #filePreOpen "cIsoLevels.exitIso()" id:(cIsoLevels.tRandom as name)
	callbacks.addScript #systemPreReset "cIsoLevels.exitIso()" id:(cIsoLevels.tRandom as name)
	callbacks.addScript #preSystemShutdown "cIsoLevels.exitIso()" id:(cIsoLevels.tRandom as name)

	--callbacks.addScript #sceneXrefPostMerge "cIsoLevels.updateXrefObjRend()" id:(cIsoLevels.tRandom as name)
	callbacks.addScript #sceneNodeAdded "cIsoLevels.sceneNodeAddedCallback()" id:(cIsoLevels.tRandom as name)
	callbacks.addScript #nodePreDelete "cIsoLevels.sceneNodeDeletedCallback()" id:(cIsoLevels.tRandom as name)